/*************************************************
 * CAMELOID Secure End-to-End Telephony System
 * Copyright (C) 2005  Dr. Tek                     
 ************************************************/

#include "cameloid.h"
#include "cameloid_protocol.h"
#include "cameloid_voice.h"
#include "cameloid_memory.h"
#include "cameloid_functions.h"



static u8
cameloid_algorithms[CAMELOID_ALGO_COUNT + 1][10] =
{"None", "Py","Phelix","Serpent","Twofish","CAST6","Rijndael"}; 


__inline__
int cameloid_name_to_algo(u8 *name)
{
	register u8 i;

	for(i = CAMELOID_ALGO_COUNT; i > 0; --i)
	{	
		if(!strcasecmp(cameloid_algorithms[i],name))
		{
			return i;
		}
	}

	return -1; 
}

__inline__
u8 * cameloid_algo_to_name(int num)
{
	return((u8 *)cameloid_algorithms[num]);
}

__inline__
int cameloid_open_random(void)
{
        register int fd;

        if((fd = open(RANDOM_DEVICE,O_RDONLY | O_NONBLOCK))<0)
        {
                cameloid_print_error("Cannot open random device");
                return -1;
        }

        return fd;

}

__inline__
int cameloid_open_urandom(void)
{
        register int fd;

        if((fd = open(URANDOM_DEVICE,O_RDONLY | O_NONBLOCK))<0)
        {
                cameloid_print_error("Cannot open random device");
                return -1;
        }

        return fd;

}

__inline__
void cameloid_get_r_random_bytes(int fd, u8 *data, u32 size)
{
	while((read(fd,data,size))<size)
	{
		continue;
	}
}


__inline__
void cameloid_get_p_random_bytes(int fd, u8 *data, u32 size)
{
        cameloid_get_r_random_bytes(fd,data,size);
	
	/* if 'use_strong_rand' is set then encrypt output of our	*/
	/* random device with rijndael for better quality random number */
	/* generation  							*/

	if(use_strong_rand)
		cameloid_get_rng_bytes(fd,data,size);
	
	return;

}

/* These are used just to obfuscate the username */

__inline__
void cameloid_add_gf256(u8 *x, u8 *y, u32 size)
{
	register u32 i = 0;

	for(i=size;i>0;--i)
	{
		x[i-1] ^= y[i-1];
	}

	return;
}


/*
 *
 *	Function to generate SHA-256 hash
 *
 *
 */

__inline__ 
void cameloid_generate_hash(const u8 *data, u32 size, u8 *hash)
{
        sha256_context ctx;       
  
        sha256_starts(&ctx);
        sha256_update(&ctx,(u8 *) data,size);
        sha256_finish(&ctx,hash);

} 

/*
 *
 *	Function to generate Tiger-192 hash
 *
 *
 */

__inline__
void cameloid_generate_hash_tiger(const u8 *data, u32 size, u8 *hash)
{
	tiger((u64 *) data,size,(u64 *) hash);
}

__inline__
u8 * cameloid_hash_and_show(const u8 *data, u32 size)
{
        register unsigned int i = 0, x = 0;
        static u8 hash[32], buf[64 + 1];
        
	cameloid_destroy_memory(hash,32);
	cameloid_destroy_memory(buf,64);
        cameloid_generate_hash(data,size,hash);
 
        for(i = 0; i < 32; i++)
        {  
                sprintf((u8 *)buf+x,"%02X",hash[i]);
                x+=2;
        }
        
        return buf;
        
}

__inline__
u8 * cameloid_hash_and_show_md5(const u8 *data, u32 size)
{
	register unsigned int i = 0, x = 0;
        static u8 buf[32 + 1];
	MD5_CTX md5ctx;

	memset(&md5ctx,0,sizeof(MD5_CTX));

	MD5Init(&md5ctx);
	MD5Update(&md5ctx,data,size);
	MD5Final(&md5ctx);
      
        memset(buf,0,33);                
        for(i = 0; i < 16; i++)
        {
                sprintf((u8 *)buf+x,"%02X", md5ctx.digest[i]);
                x+=2;
        }
 
        return (u8 *) buf;

}

	
/*
 *
 *	Input function for Negotiation Key
 *
 */

__inline__
int cameloid_get_nk(FILE *s, cameloid_mstruct_t *m)
{
	int err_cnt = 0;
	struct termios term,tsave;
	u8 *NK, *NK_TEMP;

	NK 		= m->NK;
	NK_TEMP		= m->NK_TEMP;
	tcgetattr(fileno(s),&tsave);
	term = tsave;
	term.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
        tcsetattr(fileno(s),TCSANOW,&term);
	
	while(err_cnt<5)
	{
		cameloid_destroy_memory(NK_TEMP,CAMELOID_MAX_NK);
		cameloid_destroy_memory(NK,CAMELOID_MAX_NK);

		printf("Enter NK: "); fflush(stdout);
		fgets(NK,CAMELOID_MAX_NK,s);		
		NK[strlen(NK) - 1] = 0;
		if(strlen(NK) < CAMELOID_MIN_NK)
		{
			printf("\nNK is too small, must be at least %3d-bits(%d bytes) long\n",CAMELOID_MIN_NK*8,CAMELOID_MIN_NK);
			fflush(stdout);
			err_cnt++;
			continue;
		}

		printf("\nEnter NK again: "); fflush(stdout);
		fgets(NK_TEMP,CAMELOID_MAX_NK,s);
		NK_TEMP[strlen(NK_TEMP) - 1] = 0;

		if(!strcmp(NK,NK_TEMP))
		{
			cameloid_destroy_memory(NK_TEMP,CAMELOID_MAX_NK);	
			break;	
		}
		else
		{
			putchar(0x0A);
			err_cnt++;
			continue;
		}
	}

	cameloid_destroy_memory(NK_TEMP,CAMELOID_MAX_NK);
	printf("\n");
	tcsetattr(fileno(s),TCSANOW,&tsave);
	if(err_cnt >= 5)
                return -1;
	
	return 0;
}



/*
 *
 *	Key Setup Function
 *
 */

__inline__
void cameloid_key_setup(int algo, const u8 *key, u32 keysize, cameloid_kc_t *ctx)
{	
	static int init_algs = 0;
	u8 fixed_key_hash[32];
	
	cameloid_destroy_memory((u8 *) ctx,sizeof(cameloid_kc_t));

	if(!init_algs)
	{
		PhelixInit();
		PyInit();
		init_algs = 1;
	}

	ctx->alg = algo;
	
	cameloid_destroy_memory(fixed_key_hash,32);
        cameloid_generate_hash(key,keysize,fixed_key_hash);
	cameloid_generate_hash_tiger(key,keysize,ctx->keydatahash);


	/* 
	 *
	 * Initiate the KSA for the selected cipher
 	 *
	 */

	switch(algo)
	{
		
		/* Py(roo) */
		case CAMELOID_ALG_PY:
			PySetupKey(&ctx->py_ctx,fixed_key_hash,256,128);
			break;

		/* Phelix */
		case CAMELOID_ALG_PHELIX:
			PhelixSetupKey(&ctx->phelix_ctx,fixed_key_hash,256,128,0);
			break;
			
		/* Serpent */
		case CAMELOID_ALG_SERPENT:
                        SerpentKeySetup(&ctx->serpent_ctx,256,fixed_key_hash);
			break;

		/* Twofish */
		case CAMELOID_ALG_TWOFISH:
                        TwofishSetupKey(&ctx->twofish_ctx,fixed_key_hash,32);
			break;

		/* CAST6 */
		case CAMELOID_ALG_CAST6:
                        CAST6SetupKey(&ctx->cast6_ctx,fixed_key_hash,32);
			break;

		/* Rijndael */
		case CAMELOID_ALG_RIJNDAEL:
			RijndaelSetupKey(&ctx->rijndael_ctx,fixed_key_hash,256);
			break;
	}
	
	cameloid_destroy_memory(fixed_key_hash,32);
	return;
}


/*
 *
 *	Encryption/Decryption Routines
 *
 */

__inline__
void cameloid_crypt_data(u8 mode, u8 *data, u32 size, u8 *real_nonce, cameloid_kc_t *ctx)
{
	register u32 i = 0, n = 0, cnt = 0;
	u8 iv_hash_output[25], nonce[ 24  + CAMELOID_NONCE_SIZE + 1 ];
	int algo = ctx->alg;


	/* We do this because the encryption/decryption routines for the */
	/* the block ciphers are the same in counter mode		 */

	if(algo == CAMELOID_ALG_SERPENT || 
	   algo == CAMELOID_ALG_TWOFISH || 
	   algo == CAMELOID_ALG_CAST6	||
	   algo == CAMELOID_ALG_RIJNDAEL) 
	   {
		mode = CAMELOID_MODE_ENCRYPT;
	   }


	/* Here we take in nonce (real_nonce) and concantenate it with */
	/* the tiger-192 hash of our key (ctx->keydatahash) and then   */
	/* we hash it then truncate it to make our 128-bit nonce */
 
	cameloid_destroy_memory(nonce,24 + 32 + 1);
	memcpy(nonce, real_nonce, CAMELOID_NONCE_SIZE);
	memcpy(nonce + CAMELOID_NONCE_SIZE, ctx->keydatahash, 24);
	cameloid_generate_hash_tiger(nonce,CAMELOID_NONCE_SIZE + 24,iv_hash_output);
	

	/* Encryption/decryption */
	switch(mode)
	{
		case CAMELOID_MODE_ENCRYPT:

	
			/* Py(roo) encryption routine */
			if(algo == CAMELOID_ALG_PY)
			{
				PySetupNonce(&ctx->py_ctx,iv_hash_output);
				PyEncryptBytes(&ctx->py_ctx,data,data,size);
			}

			/* Phelix encryption routine */
			if(algo == CAMELOID_ALG_PHELIX)
			{
				PhelixSetupNonce(&ctx->phelix_ctx,iv_hash_output);
				PhelixEncryptBytes(&ctx->phelix_ctx,data,data,size);
			}		
				
			
			/* Serpent counter-mode encryption/decryption routine */
			if(algo == CAMELOID_ALG_SERPENT)   
                        {
                                
                                 for(cnt = 0, i = size; i > 0; --i, cnt++)
                                 {
                                         if((cnt % 16) == 0)
                                         {
                                                 SerpentEncrypt((u32 *)iv_hash_output,(u32 *)iv_hash_output, ctx->serpent_ctx.subkeys);
                                                 for(n = 16; n > 0; --n)
                                                 {
                                                        iv_hash_output[n-1]++;
                                                        if(iv_hash_output[n-1] != 0)
                                                                break;
                                                 }
                                        }
                                        data[i-1] ^= iv_hash_output[cnt % 16];
                                }
 
                        }
                         
			/* Twofish counter-mode encryption/decryption routine */
			if(algo == CAMELOID_ALG_TWOFISH)  
                        {
                                          
                                 for(cnt = 0, i = size; i > 0; --i, cnt++)
                                 {
                                         if((cnt % 16) == 0)
                                         {
                                                 TwofishEncrypt(&ctx->twofish_ctx,(u8 *)iv_hash_output,(u8 *)iv_hash_output);
                                                 for(n = 16; n > 0; --n)
                                                 {
                                                        iv_hash_output[n-1]++;
                                                        if(iv_hash_output[n-1] != 0)
                                                                break;
                                                 }
                                        }
                                        data[i-1] ^= iv_hash_output[cnt % 16];
                                }
                          
                        }
                             
			/* CAST6 counter-mode encryption/decryption routine */
			if(algo == CAMELOID_ALG_CAST6)
                        {                
                               
                                 for(cnt = 0, i = size; i > 0; --i, cnt++)
                                 {
                                         if((cnt % 16) == 0)
                                         {
                                                 CAST6Encrypt(&ctx->cast6_ctx,(u8 *)iv_hash_output,(u8 *)iv_hash_output);
                                                 for(n = 16; n > 0; --n)
                                                 {
                                                        iv_hash_output[n-1]++;
                                                        if(iv_hash_output[n-1] != 0)
                                                                break;
                                                 }
                                        }
                                        data[i-1] ^= iv_hash_output[cnt % 16];
                                }
                                
                        }

			/* Rijndael counter-mode encryption/decryption routine */
                        if(algo == CAMELOID_ALG_RIJNDAEL)
                        {
                                          
                                 for(cnt = 0, i = size; i > 0; --i, cnt++)
                                 {
                                         if((cnt % 16) == 0)
                                         {
                                                 RijndaelEncrypt(&ctx->rijndael_ctx,(u8 *)iv_hash_output,(u8 *)iv_hash_output);
                                                 for(n = 16; n > 0; --n)
                                                 {
                                                        iv_hash_output[n-1]++;
                                                        if(iv_hash_output[n-1] != 0)
                                                                break;
                                                 }
                                        }
                                        data[i-1] ^= iv_hash_output[cnt % 16];
                                }

                        }

			break;

			
		case CAMELOID_MODE_DECRYPT:
			
			/* Py(roo) decryption routine */
			if(algo == CAMELOID_ALG_PY)
                        {
			        PySetupNonce(&ctx->py_ctx,iv_hash_output);
                                PyDecryptBytes(&ctx->py_ctx,data,data,size);
                        }

			/* Phelix decryption routine */
                        if(algo == CAMELOID_ALG_PHELIX)
                        {
                                PhelixSetupNonce(&ctx->phelix_ctx,iv_hash_output);
                                PhelixDecryptBytes(&ctx->phelix_ctx,data,data,size);
                        }
			break;

	}

	return;

}


__inline__
u8 cameloid_get_random_algo(void)
{
	u8  n = 0;
	u32 v = 0;

	/* Does this pass DIEHARD ;) */

	srandom(time(NULL) * getpid());
	while(!n)
	{
		v = random();
		n = ((v>>0 & 0xFF) + (v>>16 & 0xFF) + (v>>24 & 0xFF) + (v>>8 &  0xFF)) % CAMELOID_ALGO_COUNT+1; 
		srandom(v);
	}
	return n;
}
